{$mode delphi}
unit DIB16;

interface

uses
  Windows,
  Messages,
//  WinProcs,
//  WinTypes,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  stdctrls,
  Math,
  ExtCtrls;

type

TDIB16bit = class(Tobject)
  private
    {Privates}
    FDibHandle : HBitmap;  {Bitmap Handle to the DIB}
    FBheader : TBitMapInfo; {Holds info about the DIB}
    FPointerToBitmap : Pointer; {The pointer to the DIB's memory}
    FScanWidth : Integer; {The width in bytes of a scanline}
    FDeviceContext : HDC; {A device context for the DIB}
    FScanWidthArray : array [0..1023] of Integer;  {starting address of each scanline}
    FZValues: Array[0..767,0..1023] Of Single; {keeping the Z-depth of drawn pixels}
    function GetHeight : Integer;
    function GetWidth : Integer;
  protected
    { Protected declarations }
  public
    { Public declarations }
    procedure FlipBackPage(DeviceContext : HDC);
    procedure SetPixel(X, Y : Integer; Color : Word);
    function GetPixel(X, Y : Integer) : Word;
    Function GetPixelZ(X,Y: Integer): Single;
    Procedure SetPixelZ(X,Y: Integer; Z: Single);
    procedure ClearBackPage(Color : Word);
    Procedure ClearZValues;
    procedure DrawHorizontalLine (Y, X1, X2 : Integer; Z1,Z2: Single; Color : Word; Fill: Boolean) ;
    function GetHandle : HBitmap; {Returns the Bitmap handle of the DIB}
    function GetDC : HDC;
    constructor Create(Height, Width : Integer);
    destructor Destroy; override;
    property Width : Integer read GetWidth;
    property Height : Integer read GetHeight;
  end;

implementation

procedure TDIB16bit.FlipBackPage(DeviceContext : HDC);
begin
{This copies the DIB to the device context (usually a handle to a canvas)
 specified in the parameter DeviceContext : HDC}
  StretchDIBits(DeviceContext, 0, 0, FBHeader.bmiHeader.biWidth,
   FBHeader.bmiHeader.biHeight, 0, 0, FBHeader.bmiHeader.biWidth,
   FBHeader.bmiHeader.biHeight, FPointerToBitmap, FBHeader, DIB_RGB_COLORS, SRCCOPY);
end;

function TDIB16bit.GetHeight : Integer;
begin
     result := FBHeader.bmiHeader.biHeight;
end;

function TDIB16bit.GetWidth : Integer;
begin
     result := FBHeader.bmiHeader.biWidth;
end;

function TDIB16bit.GetDC : HDC;
begin
     result := FDeviceContext;
end;

function TDIB16bit.GetPixel(X, Y : Integer) : Word;
var
  BasePointer : ^Word;
begin
{Assigns BasePointer the memory address of the pixel in the DIB by adding to the
beginning of the address (FPointerBitmap) the Y value multiplied by the ScanWidth.  Then the X is added
to the result.  X is multiplied by two because a pixel in 16bit color takes up 2 bytes
of memory}

//*** Optimization shift multiplication - Remove the * 2 and make it shl 1 ***
//*** Optimization Look Up Table - use the FScanWidthArray for precalculated
//*** FScanwidth * Y + FPointerToBitmap

  BasePointer := Pointer(FScanWidthArray[Y] + (X shl 1));
  result := BasePointer^;
end;


procedure TDIB16bit.SetPixel(X, Y : Integer; Color : Word);
var
  BasePointer : ^Word;
begin
{Assigns BasePointer the memory address of the pixel in the DIB by adding to the
beginning of the address (FPointerBitmap) the Y value multiplied by the ScanWidth.  Then the X is added
to the result.  X is multiplied by two because a pixel in 16bit color takes up 2 bytes
of memory}

//*** Optimization shift multiplication - Remove the * 2 and make it shl 1 ***
//*** Optimization Look Up Table - use the FScanWidthArray for precalculated
//*** FScanwidth * Y + FPointerToBitmap

  BasePointer := Pointer(FScanWidthArray[Y] + (X shl 1));
  BasePointer^ := Color;
end;

procedure TDIB16bit.ClearBackPage(Color : Word);
var
  X : Integer;
  BasePointer : ^Integer;
  XColor : Integer;
begin
{Loops through from the beginning of the bitmap to the end setting every pixel
to the 'Color' parameter}

  BasePointer := FPointerToBitmap;

  //*** Optimization writing 2 pixels at a time - Move the Word Color into the  ***
  //*** top 16bits and the bottom 16bits of an 32bit Integer and blit 32bits at ***
  //*** a time rather than 16bit (half the number of loops required to clear    ***
  //*** the back page ***

  XColor := Color or (color shl 16); {Changed from shr 16 - a pervious error}

  //*** Optimization shift division - Remove the div 2 and make it shr 1 ***
  for X := 0 to (((FScanWidth shr 1) * (FBHeader.bmiHeader.biHeight)) - 1) shr 1 do
    begin
      BasePointer^ := XColor;
      {inc increments a pointer by the size of the type that the pointer points to.
      In other words 4 bytes}
      inc(BasePointer);
    end;
end;

Function TDIB16bit.GetPixelZ(X,Y: Integer): Single;
  Begin
  Result:=FZValues[Y,X];
  End;   {GetPixelZ}


Procedure TDIB16bit.SetPixelZ(X,Y: Integer; Z: Single);
  Begin
  FZValues[Y,X]:=Z;
  End;


procedure TDIB16bit.ClearZValues;
  Var
    I: Integer;
    J: integer;
  Begin
  For I:=0 To 767 Do
    Begin
    For j:=0 To 1023 Do
      Begin
      FZValues[i,j]:=-1.0e+38;
      End;
    End;
  End;


procedure TDIB16bit.DrawHorizontalLine (Y, X1, X2 : Integer; Z1,Z2: Single; Color : Word; Fill: Boolean) ;
var
  X : Integer;
  BasePointer : ^Word;
  Z,Zincr: Single;

begin
{Assigns BasePointer the memory address of the pixel in the DIB by adding to the
beginning of the address (FPointerBitmap) the Y value multiplied by the ScanWidth.  Then the X1 is added
to the result.  X1 is multiplied by two becasue a pixel in 16bit color takes up 2 bytes
of memory.  Then the code just loops through the number of pixels in the line}

//*** Optimization Look Up Table - use the FScanWidthArray for precalculated
//*** FScanwidth * Y + FPointerToBitmap
//*** Optimization shift multiplication - Remove the * 2 and make it shl 1 ***
If (Y < 0) Or (Y>767) Then
  begin
  Exit;
  End;
X1:=Max(0,Min(X1,1023));
X2:=Max(0,Min(X2,1023));
If Fill Then
  Begin
  Z:=Z1;
  If X1 = X2 Then
    begin
    ZIncr:=0;
    End
  Else
    Begin
    ZIncr:=(Z1 - Z2) / (X1 - X2);
    End;
  BasePointer := Pointer(FScanWidthArray[Y] + (X1 shl 1));
  for X := 0 to (X2 - X1) do
    begin
    If Z >= FZValues[Y,X1] Then
      Begin
      BasePointer^ := Color;
      FZValues[Y,X1] :=Z;
      End;
    Z:=Z + Zincr;
    {inc moves the pointer along 1 pixel in memory (2 bytes)}
    inc(BasePointer);
    inc(X1);
    end;
  End
Else
  Begin
  BasePointer := Pointer(FScanWidthArray[Y] + (X1 shl 1));
  If Z1 >= FZValues[Y,X1] Then
    Begin
    BasePointer^ := Color;
    FZValues[Y,X1] :=Z1;
    End;
  BasePointer := Pointer(FScanWidthArray[Y] + (X2 shl 1));
  If Z2 >= FZValues[Y,X2] Then
    Begin
    BasePointer^ := Color;
    FZValues[Y,X2] :=Z2;
    End;
  End;
end;

function TDIB16bit.GetHandle : HBitmap;
begin
  result := FDibHandle;
end;

constructor TDIB16Bit.Create(Height, Width : Integer);
var
   i : Integer;
begin
  Inherited Create;
  {Set up the FBHeader with the values it needs to know what type of DIB is required}
  FBHeader.bmiHeader.biWidth := Width;   {Width in pixels}
  FBHeader.bmiHeader.biHeight := Height; {Height in pixels}
  FBHeader.bmiHeader.biPlanes := 1;  {biPlanes is always 1}
  FBHeader.bmiHeader.biBitCount := 16; {color depth is 16-bit}
  FBHeader.bmiHeader.biCompression := BI_RGB;  {no compression}
  {The width in bytes of every scanline}
  FScanWidth := (((16 * FBHeader.bmiHeader.biWidth) + 31) div 32) * 4;
  {Work out the size of memory in bytes needed for the DIB}
  FBHeader.bmiHeader.biSizeImage := FScanWidth * Height;
  {These can be ignored and set to 0 for 16bit color}
  FBHeader.bmiHeader.biXPelsPerMeter := 0;
  FBHeader.bmiHeader.biYPelsPerMeter := 0;
  FBHeader.bmiHeader.biclrUsed := 0;
  FBHeader.bmiHeader.biclrImportant := 0;
  {The size of the bmiHeader structure}
  FBHeader.bmiHeader.biSize := 40;
  {Not used in 16bit color}
  FBHeader.bmiColors[0].rgbRed := 255;
  FBHeader.bmiColors[0].rgbBlue := 255;
  FBHeader.bmiColors[0].rgbGreen := 255;
  FBHeader.bmiColors[0].rgbReserved := 255;
  {Create a Device context that is compatable with the current screen}
  FDeviceContext := CreateCompatibleDC(0);
  {Now the DIB is allocated and it returns a HBitmap handle to itself}
  FDibHandle := CreateDibSection(FDeviceContext, FBHeader, DIB_RGB_COLORS, FPointerToBitmap, 0, 0);
  For i := 0 To 1023 Do
    Begin
    FScanWidthArray[i] := (FScanWidth * i) + Integer(FPointerToBitmap);
    End;
end;

Destructor TDIB16bit.Destroy;
  Begin
  DeleteObject(FDibHandle);  {Delete the DIB}
  DeleteDC(FDeviceContext);  {Delete the Device Context}
  Inherited;
end;

end.
