{$mode delphi}
unit gmp;

interface

uses
  Windows,
  Messages,
//  Windows, {WinProcs,}
//  WinTypes,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  stdctrls,
  ExtCtrls,
  DIB16,
  Math;

const
  {You can just change these constants instead of hunting through the code
  to restrict the number of polygons.  The MaxPoints should allways be 4*
  the number of MaxPolys}
  MAXPOLYS = 100;
  MAXPOINTS = 400;

type

TRenderMode = (rmWireframe, rmSolid, rmSolidShade, rmSolidTexture, rmShadedTexture);

TYBucket =
  Record
  StartX : Integer;
  EndX: Integer;
  StartZ: Single;
  EndZ : Single;
  End;

TTextureBucket =
  Record
  StartPosition: TPoint;
  EndPosition : TPoint;
  End;

TPoint3D =
  Record
  X: Single;
  Y: Single;
  Z: Single;
  End;

TLine3D =
  Record
  StartPoint: TPoint3D;
  EndPoint: TPoint3D;
  End;

TPolygon =
  Record
  Point : Array [0..3] of TPoint3D; {only allow polygons for now if to 4 points}
  NumberPoints : Integer;
  Visible : Boolean;
  AverageZ : Single;
  PolyColor : TColor;  {Color of the polygon}
  DibColor : Word;     {Word color value of PolyColor}
  Intensity : Byte;    {The light intensity of the polygon}
  End;

TObject3D =
  Record
  PolyStore  : Array [0..MAXPOLYS] of TPolygon; {Stores the local coordinates}
  PolyWorld  : Array [0..MAXPOLYS] of TPolygon; {Stores the world coordinates}
  PolyCamera : Array [0..MAXPOLYS] of TPolygon; {Stores the camera coordinates}
  NumberPolys : Integer; {Number of Polygons}
  Color : Tcolor; {Color of whole object - used in wireframe and solid color modes}
  DibColor : Word; {16bit color of whole object - used in wireframe and solid color modes}
  World : TPoint3D; {Position of the object in the world}
  End;

PObject3D = ^TObject3D; {Pointer to TObject3D}

TBeforeFlip = Procedure (Canvas : TCanvas) of Object; {Declaration for an event}

TBitmapStorage = Array [0..127, 0..127] of Word; {Declaration for Bitmap Storage}

TGMP = Class(TCustomControl)
  Private
    { Private declarations }
    FCurrentBitmap : TBitmapStorage; {Storage for the current bitmap}
    FDib : TDib16Bit; {The DIB class back buffer}
    FBackgroundDIB : TDib16Bit; {Stores the background bitmap}
    FBeforeFlip : TBeforeFlip; {Flip event member}
    FBackBuffer : TBitmap;  {Used to assign the FDIB handle to for use with the GDI}
    FColor : TColor;  {Used to hold a temporary color value in drawing polygons}
    FDibColor : Word; {Used to hold a temporary color value in drawing polygons}
    FIntensity : Byte; {Used to hold a temporary light intensity value in drawing polygons}
    ViewWidth : Integer;
    ViewHeight : Integer;
    FRenderMode : TRenderMode;
    HalfScreenWidth :Integer;
    HalfScreenHeight : Integer;
    ViewingDistance : Integer;
    YBuckets : Array [0..767] of TYBucket; {768 being the most we will allow the screen height to go up to}
    YBMin: Integer;
    YBMax: Integer;
    TextureBuckets : array [0..767] of TTextureBucket; {768 being the most we will allow the screen height to go up to}
    IntensityLUT : Array [0..31, 0..31] of Integer; {The light intensity lookup table}
    FLightStrength : Single;
    AmbientLight : Integer; {The amount of ambient light in the scene}
    TransLightSource : TPoint3D;
    FAnimationList : TList; {Holds a list of all the animation frames}
    FBackgroundBitmap : TPicture; {Holds the background bitmap}
    FEnableBackground : Boolean; {Wether or not the Background is enabled}
    procedure DrawLine3D(X1, Y1, Z1, X2, Y2, Z2 : Single);
    procedure DrawLine2DWireframe(X1, Y1, X2, Y2 : Integer);
    procedure DrawLine2DSolid(X1, Y1, X2, Y2 : Integer; Z1,Z2: Single);
    procedure ClearYBuckets;
    procedure RenderYBuckets ;
    procedure GetVector3D(var EndPoint, StartPoint, Vector : TPoint3D);
    procedure CrossProduct(var U, V, Normal : TPoint3D);
    procedure GetNormal(var P1, P2, P3, normal : TPoint3D);
    function VectorMagnitude(var Normal : TPoint3D) : Single;
    function DotProduct(var U, V : TPoint3D) : Single;
    procedure RemoveBackfacesAndShade(var AnObject : TObject3D);
    procedure DrawHorizontalLine (Y, X1, X2 : Integer; Z1,Z2: Single) ;
    procedure OrderZ(var Object3D : TObject3D);
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure DrawLine3DTexture(var StartPoint, EndPoint : TPoint3D; var TextStart, TextEnd : TPoint);
    procedure DrawLine2DTexture(var StartPoint, EndPoint, TextStart, TextEnd : TPoint);
    function CalculateRGBWord(Clr : TColor) : Word;
    procedure LocalToWorld(var AnObject : TObject3D);
    procedure WorldToCamera(var AnObject : TObject3D);
    function GetShadedWord (Texture : Word; Intensity : Integer) : Word;
    procedure RotatePoint(X, Y, Z, PreCalSin,PreCalCos : Single; var Point : TPoint3D);
    procedure Copybackground;
    procedure SetBackgroundBitmap ( Value : TPicture );
    function GetBackgroundBitmap : TPicture;
    procedure SetEnableBackground ( Value : Boolean );
  protected
    { Protected declarations }
    procedure Paint; Override;    
  public
    { Public declarations }
    ViewPoint : TPoint3D; {ViewPoint position}
    LightSource : TPoint3D; {LightSource position}
    CameraPosition : TPoint3D; {The camera position}
    CameraRotation : TPoint3D; {The camera rotation}
    CamSin: TPoint3D;   {precalculated sinus of camerarotation}
    CamCos: TPoint3D;   {precalculated Cosinus of camerarotation}
    BackGroundDrawing: Boolean;
    constructor Create(AOwner : TComponent) ; override;
    destructor Destroy; override;
    procedure ClearBackPage;
    procedure SetBackColor(Value : TColor);
    Procedure PrepareRender(var Object3D: TObject3D);
    procedure RenderNow(var Object3D : TObject3D; Prepared: Boolean);
    Procedure GetObjectCenter(Var Object3D: TObject3D; Var P: TPoint3D);
    procedure FlipBackPage;
    procedure Rotate(Xt, Yt, Zt, Angle : Single; var Object3D : TObject3D);
    procedure ChangeObjectColor(var Object3D : TObject3D; Clr : TColor);
    procedure SetLightSourcePosition(Position, Direction : TPoint3D);
    procedure SetCurrentBitmap(Bitmap : TBitmap);
    procedure SetCurrentBitmapWithAnimationFrame ( Frame : Integer );
    procedure DeleteAnimationFrame ( Frame : Integer );
    procedure AddAnimationFrame ( Bitmap : TBitmap );
    function LoadGeoObject(var AnObject : TObject3D; Filename : String) : Integer;
    Procedure LoadBackGround(Value: TPicture);
    Procedure SetCameraSinCos;
    Procedure DrawLine(X1,Y1,Z1,X2,Y2,Z2: Single; Clr: TColor);
    Procedure PointRotate(var P: TPoint3D);
  published
    { Published declarations }
    property Align;
    property LightStrength : Single read FLightStrength write FLightStrength;
    property BackColor : TColor read FColor write SetBackColor;
    property EnableBackgroundBitmap : Boolean read FEnablebackground write SetEnablebackground;
    property BackgroundBitmap : TPicture read GetBackgroundBitmap write SetbackgroundBitmap;
    property RenderMode : TRenderMode read FRenderMode write FRenderMode;
    property BeforeFlip : TBeforeFlip read FBeforeFlip write FBeforeFlip;
    property OnMouseMove;
    property OnMouseDown;
    property OnMouseUp;
  end;

procedure Register;

implementation

Function TGMP.GetShadedWord(Texture : Word; Intensity : Integer) : Word;
  Var
    intRed : Integer;
    intGreen : Integer;
    intBlue : Integer;
    intBitMask : Integer;
  Begin
  intBitMask := $001F;  {31 = last 5 bits for Blue}
  intBlue := Texture and intBitMask;
  intBitMask := $03E0;  {992 = middle 5 bits for Green}
  intGreen := Texture and intBitMask;
  intGreen := intGreen shr 5;
  intBitMask := $7C00;  {31744 = first 5 bits for Red}
  intRed := Texture and intBitMask;
  intRed := intRed shr 10;
  {Now get the new shades - this uses the lookup table that we worked out previously}
  intRed := IntensityLUT[intRed, Intensity];
  intGreen := IntensityLUT[intGreen, Intensity];
  intBlue := IntensityLUT[intBlue, Intensity];
  intRed := intRed shl 10;
  intGreen := intGreen shl 5;
  Result := intRed + intBlue + intGreen;
  End;

Procedure TGMP.WorldToCamera(Var AnObject : TObject3D);
  Var
    X : Integer;
  Begin
  {Loops though all of the polygons transfering them to Polycamera}
  With AnObject Do
    Begin
    For X := 0 To NumberPolys -1 Do
      Begin
      With PolyCamera[x] Do
        Begin
        NumberPoints := PolyWorld[X].NumberPoints;
        {The camera coodinates must be minused from the world coordinates}
        Point[0].X := PolyWorld[X].Point[0].X - CameraPosition.X;
        Point[0].Y := PolyWorld[X].Point[0].Y - CameraPosition.Y;
        Point[0].Z := PolyWorld[X].Point[0].Z - CameraPosition.Z;
        Point[1].X := PolyWorld[X].Point[1].X - CameraPosition.X;
        Point[1].Y := PolyWorld[X].Point[1].Y - CameraPosition.Y;
        Point[1].Z := PolyWorld[X].Point[1].Z - CameraPosition.Z;
        Point[2].X := PolyWorld[X].Point[2].X - CameraPosition.X;
        Point[2].Y := PolyWorld[X].Point[2].Y - CameraPosition.Y;
        Point[2].Z := PolyWorld[X].Point[2].Z - CameraPosition.Z;
        If NumberPoints = 4 Then
          Begin
          Point[3].X := PolyWorld[X].Point[3].X - CameraPosition.X;
          Point[3].Y := PolyWorld[X].Point[3].Y - CameraPosition.Y;
          Point[3].Z := PolyWorld[X].Point[3].Z - CameraPosition.Z;
          End;
       {Now Rotate the points using the camera rotation}
        RotatePoint( 1, 0, 0, CamSin.X, CamCos.X, Point[0]);
        RotatePoint( 0, 1, 0, CamSin.Y, CamCos.Y, Point[0]);
        RotatePoint( 0, 0, 1, CamSin.Z, CamCos.Z, Point[0]);
        RotatePoint( 1, 0, 0, CamSin.X, CamCos.X, Point[1]);
        RotatePoint( 0, 1, 0, CamSin.Y, CamCos.Y, Point[1]);
        RotatePoint( 0, 0, 1, CamSin.Z, CamCos.Z, Point[1]);
        RotatePoint( 1, 0, 0, CamSin.X, CamCos.X, Point[2]);
        RotatePoint( 0, 1, 0, CamSin.Y, CamCos.Y, Point[2]);
        RotatePoint( 0, 0, 1, CamSin.Z, CamCos.Z, Point[2]);
        If NumberPoints = 4 Then
          Begin
          RotatePoint( 1, 0, 0, CamSin.X, CamCos.X, Point[3]);
          RotatePoint( 0, 1, 0, CamSin.Y, CamCos.Y, Point[3]);
          RotatePoint( 0, 0, 1, CamSin.Z, CamCos.Z, Point[3]);
          End;
        {Now Rotate the LightSource!}
        TransLightSource := LightSource;
        Rotatepoint( 1, 0, 0, CamSin.X, CamCos.X, TransLightSource);
        Rotatepoint( 0, 1, 0, CamSin.Y, CamCos.Y, TransLightSource);
        Rotatepoint( 0, 0, 1, CamSin.Z, CamCos.Z, TransLightSource);
        PolyColor := PolyWorld[X].PolyColor;
        DibColor := PolyWorld[X].DibColor;
        end;
      end;
    end;
  end;


Procedure TGMP.LocalToWorld(Var AnObject : TObject3D);
  Var
    X : Integer;
  Begin
  With AnObject Do
    Begin
    For X := 0 To NumberPolys -1 Do
      Begin
      With PolyWorld[x] Do
        Begin
        NumberPoints := PolyStore[X].NumberPoints;
        Point[0].X := PolyStore[X].Point[0].X + World.X;
        Point[0].Y := PolyStore[X].Point[0].Y + World.Y;
        Point[0].Z := PolyStore[X].Point[0].Z + World.Z;
        Point[1].X := PolyStore[X].Point[1].X + World.X;
        Point[1].Y := PolyStore[X].Point[1].Y + World.Y;
        Point[1].Z := PolyStore[X].Point[1].Z + World.Z;
        Point[2].X := PolyStore[X].Point[2].X + World.X;
        Point[2].Y := PolyStore[X].Point[2].Y + World.Y;
        Point[2].Z := PolyStore[X].Point[2].Z + World.Z;
        If NumberPoints = 4 Then
          Begin
          Point[3].X := PolyStore[X].Point[3].X + World.X;
          Point[3].Y := PolyStore[X].Point[3].Y + World.Y;
          Point[3].Z := PolyStore[X].Point[3].Z + World.Z;
          End;
        PolyColor := PolyStore[X].PolyColor;
        DibColor := PolyStore[X].DibColor;
        End;
      End;
    End;
  End;

function TGMP.LoadGeoObject(var AnObject : TObject3D; Filename : String) : Integer;
{function accepts a TObject3D variable and a filename and returns a integer succes/error code}
var
  GeoFile : System.Text; {Geo text file}
  Counter, X, Int1, Int2, Int3, Int4, Int5 : Integer; {Integer variables for reading in data}
  TextLine : String; {String variable for reading in a whole line of text}
  PointArray : array [0..MAXPOINTS] of TPoint3D; {Temp storage for reading in vertices up to MAXPOINTS = MAXPOLYS * 4}
begin
  {Try to open the geo file - if it doesn't exist send back an error code}
  try
    AssignFile(GeoFile, Filename);
    except
      Result := -1; {-1 means file doesn't exist}
      exit;
    end;
  Reset(GeoFile); {Sets the file up for reading}

  {Just read in the header and make sure this is a geo file}
  ReadLn(GeoFile, TextLine);
  if TextLine <> '3DG1' then
    begin
      Result := -2; {-2 means file is not a valid GEO file}
      exit;
    end;

  {Now read in the number of Vertices into the variable counter}
  ReadLn(GeoFile, Counter);

  {loop through all the vertices and store them in a temporary array}
  for X := 0 to Counter - 1 do
    begin
      Readln(GeoFile, PointArray[x].x, PointArray[x].y, PointArray[x].z);
    end;

  {Now insert the polygons into the Object and keep a count of how many there are}
  Counter := 0;
  while not EOF(GeoFile) do
    begin
      {Read in the polygon line - this tells us what vertices to join up and how many vertices there are in the polygon}
      Readln(GeoFile, int1, int2, int3, int4, int5);
      with AnObject.PolyStore[Counter] do
        begin
          {Point 1 in the polygon}
          Point[0].X := PointArray[int2].X;
          Point[0].Y := PointArray[int2].Y;
          Point[0].Z := PointArray[int2].Z;
          {Point 2 in the polygon}
          Point[1].X := PointArray[int3].X;
          Point[1].Y := PointArray[int3].Y;
          Point[1].Z := PointArray[int3].Z;
          {Point 3 in the polygon}
          Point[2].X := PointArray[int4].X;
          Point[2].Y := PointArray[int4].Y;
          Point[2].Z := PointArray[int4].Z;
          if int1 = 4 then
            begin
              {Point 4 in the polygon}
              Point[3].X := PointArray[int5].X;
              Point[3].Y := PointArray[int5].Y;
              Point[3].Z := PointArray[int5].Z;
              {Record the number of points in this polygon}
              NumberPoints := 4;
            end
          else
            {Record the number of points in this polygon}
            NumberPoints := 3;
      end;
    {Increment counter to keep track of how many polygons we have}
    inc(Counter);
    end;

  {Record the number of Polygons}
  AnObject.NumberPolys := Counter;
  {Close the file}
  CloseFile(GeoFile);

  {Return Success}
  Result := 1;
end;

Procedure TGMP.LoadBackGround(Value: TPicture);
  Begin
  SetBackGroundBitMap(Value);
  End;  {TGMP.LoadBackGround}

Procedure TGMP.SetCameraSinCos;
  Begin
  CamSin.X:=Sin(-CameraRotation.X);
  CamCos.X:=Cos(-CameraRotation.X);
  CamSin.Y:=Sin(-CameraRotation.Y);
  CamCos.Y:=Cos(-CameraRotation.Y);
  CamSin.Z:=Sin(-CameraRotation.Z);
  CamCos.Z:=Cos(-CameraRotation.Z);
  End;

Procedure TGMP.DrawLine(X1,Y1,Z1,X2,Y2,Z2: Single; Clr: TColor);
  Var
    Point1: TPoint3D;
    Point2: TPoint3D;
    Point: Tpoint3D;
    R: SIngle;
  Begin
  With Point1 Do
    Begin
    X:=X1-CameraPosition.X;
    Y:=Y1-CameraPosition.Y;
    Z:=Z1-CameraPosition.Z;
    End;
  With Point2 Do
    Begin
    X:=X2-CameraPosition.X;
    Y:=Y2-CameraPosition.Y;
    Z:=Z2-CameraPosition.Z;
    End;

  Point:=Point1;

  Point1.Y := Point.Y * CamCos.X - Point.Z * CamSin.X;
  Point1.Z := Point.Y * CamSin.X + Point.Z * CamCos.X;
  Point:=Point1;
  Point1.Z := Point.Z * CamCos.Y - Point.X * CamSin.Y;
  Point1.X := Point.Z * CamSin.Y + Point.X * CamCos.Y;
  Point:=Point1;
  Point1.X := Point.X * CamCos.Z - Point.Y * CamSin.Z;
  Point1.Y := Point.X * CamSin.Z + Point.Y * CamCos.Z;

  Point:=Point2;

  Point2.Y := Point.Y * CamCos.X - Point.Z * CamSin.X;
  Point2.Z := Point.Y * CamSin.X + Point.Z * CamCos.X;
  Point:=Point2;
  Point2.Z := Point.Z * CamCos.Y - Point.X * CamSin.Y;
  Point2.X := Point.Z * CamSin.Y + Point.X * CamCos.Y;
  Point:=Point2;
  Point2.X := Point.X * CamCos.Z - Point.Y * CamSin.Z;
  Point2.Y := Point.X * CamSin.Z + Point.Y * CamCos.Z;

  If (Point1.Z >=0) And (Point2.Z>=0) Then
    Begin
    Exit;
    End;

  If Point2.Z > 0 Then
    Begin
    R:= (Point1.Z - 1) / (Point1.Z - Point2.Z);
    Point2.X:=Point1.X + (Point2.X - Point1.X) * R;
    Point2.Y:=Point1.Y + (Point2.Y - Point1.Y) * R;
    Point2.Z:=-1;
    End;
  If Point1.Z > 0 Then
    Begin
    R:= (Point2.Z - 1) / (Point2.Z - Point1.Z);
    Point1.X:=Point2.X + (Point1.X - Point2.X) * R;
    Point1.Y:=Point2.Y + (Point1.Y - Point2.Y) * R;
    Point1.Z:=-1;
    End;


  FBackBuffer.Canvas.Pen.Color := Color;
  DrawLine3D(Point1.X,Point1.Y,Point1.Z,Point2.X,Point2.Y,Point2.Z);
  End;

Procedure TGMP.PointRotate(var P: TPoint3D);
  Begin
  RotatePoint( 1, 0, 0, CamSin.X, CamCos.X, P );
  RotatePoint( 0, 1, 0, CamSin.Y, CamCos.Y, P );
  RotatePoint( 0, 0, 1, CamSin.Z, CamCos.Z, P );
  End;  {TGMP.PointRotate}

function TGMP.CalculateRGBWord(Clr : TColor) : Word;
var
  Calc : Single;
  R,G,B : Integer;
begin
  {GetRValue, GetGValue, GetBValue return a value which is based on the color scale of  0 to 255}

  {This gets the red value and rescales it from a 1/256 number to a 0/31 number}
  {Use Recipricals where possible - 1/2.56 :=  0.390625}
  Calc := GetRValue(Color) * 0.390625;
  Calc := Calc * 0.31;
  R := Round(Calc);

  {This gets the green value and rescales it from a 1/256 number to a 0/31 number}
  {Use Recipricals where possible - 1/2.56 :=  0.390625}
  Calc := GetGValue(Color) * 0.390625;
  Calc := Calc * 0.31;
  G := Round(Calc);

  {This gets the red value and rescales it from a 1/256 number to a 0/31 number}
  {Use Recipricals where possible - 1/2.56 :=  0.390625}
  Calc := GetBValue(Color) * 0.390625;
  Calc := Calc * 0.31;
  B := Round(Calc);

  {The B Value is the last 5 bits so it can be left alone}
  {R must be shifted left by 10 bits so that it sits at the position 15-11}
  R := R shl 10;
  {G must be shifted to the left 5 bits so that it sits at the position 10..6}
  G := G shl 5;
  {Then just add them together}
  Result := R + G + B;
end;

procedure TGMP.DrawLine3DTexture(var StartPoint, EndPoint : TPoint3D; var TextStart, TextEnd : TPoint);
var
  NewStartPoint, NewEndPoint : TPoint;
begin
  {We no longer need to have the property Z distance because that is all included in the
  world coordinate system}
  NewStartPoint.X := HalfScreenWidth - Round(StartPoint.X * ViewingDistance / (StartPoint.Z));
  NewStartPoint.Y := Round(HalfScreenHeight - StartPoint.Y * ViewingDistance / (StartPoint.Z));
  NewEndPoint.X := HalfScreenWidth - Round(EndPoint.X * ViewingDistance / (EndPoint.Z));
  NewEndPoint.Y := Round(HalfScreenHeight - EndPoint.Y * ViewingDistance / (EndPoint.Z));
  case RenderMode of
    rmSolidTexture : DrawLine2DTexture(NewStartPoint, NewEndPoint, TextStart, TextEnd);
    rmShadedTexture : DrawLine2DTexture(NewStartPoint, NewEndPoint, TextStart, TextEnd);
  end;
end;


procedure TGMP.DrawLine2DTexture(var StartPoint, EndPoint, TextStart, TextEnd : TPoint);
var
  CurrentX, XIncr : single;
  TextX, TextY, TextXIncr, TextYIncr : Integer;
  Y , Length : Integer;
  TempPoint : TPoint;
begin
  {No point in drawing horizontal lines! The rest of the polygon will define the edges}
  if StartPoint.Y = EndPoint.Y then exit;

  {Swap if Y1 is less than Y2 so we are always drawing from top to bottom}
  if EndPoint.Y < StartPoint.Y then
    begin
      TempPoint.X := StartPoint.X;
      TempPoint.Y := StartPoint.Y;
      StartPoint.X  := EndPoint.X;
      StartPoint.Y  := EndPoint.Y;
      EndPoint.X  := TempPoint.X;
      EndPoint.Y  := TempPoint.Y;
      TempPoint.X := TextEnd.X;
      TempPoint.Y := TextEnd.Y;
      TextEnd.X := TextStart.X;
      TextEnd.Y := TextStart.Y;
      TextStart.X := TempPoint.X;
      TextStart.Y := TempPoint.Y;
    end;

  Length := EndPoint.Y - StartPoint.Y;

  {Xincr is how much the X must increment though each Y increment.}
  XIncr := (EndPoint.X - StartPoint.X) / Length;
  CurrentX := StartPoint.X;

  {Work out the TextX Increment and TextY Increment}
  TextXIncr := ((TextEnd.X - TextStart.X) shl 8) div Length;
  TextYIncr := ((TextEnd.Y - TextStart.Y) shl 8) div Length;
  TextX := TextStart.X shl 8;
  TextY := TextStart.Y shl 8;

  {Now to loop through the Y Values and fill up the Y buckets}
  for Y := StartPoint.y to EndPoint.y do
    begin
      {Basic Y Clipping}
      if (Y < 0) or (Y > ViewHeight-1) then
        begin
          TextX := TextX + TextXIncr;
          TextY := TextY + TextYIncr;
          CurrentX := CurrentX + XIncr;
          Continue;
        end;

      {All Y buckets are initialized to -16000}
      if YBuckets[Y].StartX = -16000 then
        begin
	  YBuckets[Y].StartX := Round(CurrentX);
	  YBuckets[Y].EndX := Round(CurrentX);
          TextureBuckets[Y].StartPosition.X := TextX shr 8;
          TextureBuckets[Y].StartPosition.Y := TextY shr 8;
          TextureBuckets[Y].EndPosition.X := TextX shr 8;
          TextureBuckets[Y].EndPosition.Y := TextY shr 8;
	end
      else
        begin
          if CurrentX > YBuckets[Y].EndX then
            begin
              YBuckets[Y].EndX := Round(CurrentX);
              TextureBuckets[Y].EndPosition.X := TextX shr 8;
              TextureBuckets[Y].EndPosition.Y := TextY shr 8;
            end;
          if CurrentX < YBuckets[Y].StartX then
            begin
              YBuckets[Y].StartX := Round(CurrentX);
              TextureBuckets[Y].startPosition.X := TextX shr 8;
              TextureBuckets[Y].startPosition.Y := TextY shr 8;
            end;
	end;
      CurrentX := CurrentX + XIncr;
      TextX := TextX + TextXIncr;
      TextY := TextY + TextYIncr;
    end;
end;

procedure TGMP.Paint;
var
   tmpBitmap : TBitmap;
begin
  inherited;
  {New ClearBackPage using the DIB class' methods.  Notice that it now passed
  a 16bit color value rather than a TColor}
  FDib.ClearBackPage(CalculateRGBWord(FColor));
  {Create a bitmap}
  tmpBitmap := TBitmap.Create;
  {Assign the background's Handle to it}
  tmpBitmap.Handle := FbackgroundDIB.GetHandle;

  {Stetch it to the full window}
  tmpBitmap.Canvas.StretchDraw ( ClientRect, backgroundBitmap.Bitmap );

  {Release the handle and free the temporary bitmap}
  tmpBitmap.ReleaseHandle;
  tmpBitmap.Free;
  {Copy background buffer to main screen}
  FlipBackPage;
end;

procedure TGMP.SetBackColor(Value : TColor);
begin
  FColor := Value;
  Paint;
end;

procedure TGMP.WMSize(var Message: TWMSize);
begin
  Inherited;
  {Check height does not exceed our allow maximim}
  If Height > 768 then Height := 768;

  {Get the height and width of the window}
  ViewHeight := Height;
  ViewWidth := Width;

  {set the bitmaps height}
  FBackBuffer.Height := ViewHeight;
  FBackBuffer.Width := ViewWidth;

  {Set up viewport}
  HalfScreenHeight := ViewHeight div 2;
  HalfScreenWidth := ViewWidth div 2;

  {Free's the current DIB}
  FDib.Free;
  FbackgroundDIB.Free;

  {Set's it up with the new Height and Width}
  FDib := TDib16Bit.Create(ViewHeight, ViewWidth);
  FBackgroundDIB := TDib16Bit.Create(ViewHeight, ViewWidth);

  {Set FDib's bitmap handle to a FBackBuffer for use with the
  BeforeFlip event}
  FBackBuffer.Handle  := FDib.GetHandle;
  Paint;
end;

procedure TGMP.SetCurrentBitmap(Bitmap : TBitmap);
var
  X, Y : Integer;
begin
  {The FCurrentBitmap now holds the coverted 16bit values rather
  than the TColor values}
  for X := 0 to 127 do
    for Y := 0 to 127 do
      FCurrentBitmap[x,127-y] := CalculateRGBWord(Bitmap.Canvas.Pixels[X,Y]);
end;

procedure TGMP.SetCurrentBitmapWithAnimationFrame ( Frame : Integer );
begin
     {Make sure the frame is a valid one}
     If (Frame < 0) or (Frame > FAnimationList.Count) then exit;
     {Copy the memory of the animation frame into FCurrentBitmap}
     CopyMemory( @FCurrentBitmap, FAnimationList[Frame], sizeof(TBitmapStorage) );
end;

procedure TGMP.AddAnimationFrame ( Bitmap : TBitmap );
var
   X, Y : Integer;
   pBitmapStorage : ^TBitmapStorage;
begin
     {Create a dynamic TBitmapStorage}
     New ( pBitmapStorage );
     {Add it to the animation list}
     FAnimationList.Add ( pBitmapStorage );
     {Fill in the data}
     for X := 0 to 127 do
         for Y := 0 to 127 do
         pBitmapStorage^[x,127-y] := CalculateRGBWord(Bitmap.Canvas.Pixels[X,Y]);
end;

procedure TGMP.DeleteAnimationFrame ( Frame : Integer );
var
   pBitmapStorage : ^TBitmapStorage;
begin
     {Check to see the it is a valid frame}
     If (Frame < 0) or (Frame > FAnimationList.Count) then exit;
     {Get a pointer to the TBitmapStorage}
     pBitmapStorage := FAnimationList[Frame];
     {Dispose of the memory}
     Dispose(pBitmapStorage);
     {Decrease the list}
     FAnimationList.Delete(Frame);
end;


procedure TGMP.OrderZ(var Object3D : TObject3D);
var
  X, Y : Integer;
  Temp : TPolygon;
begin
  {new optimizations}
  With Object3D do
  begin
       for Y := 0 to NumberPolys - 1 do
       begin
            if PolyCamera[Y].NumberPoints = 3 then
            PolyCamera[Y].AverageZ := (PolyCamera[Y].Point[0].Z + PolyCamera[Y].Point[1].Z +
                                           PolyCamera[Y].Point[2].Z) * 0.333;

            if PolyCamera[Y].NumberPoints = 4 then
            PolyCamera[Y].AverageZ := (PolyCamera[Y].Point[0].Z + PolyCamera[Y].Point[1].Z +
                                           PolyCamera[Y].Point[2].Z + PolyCamera[Y].Point[3].Z) * 0.25;
       end;

       {A very simple way to speed this up would be to use a different sorting algorithm - for instance quicksort}
       for X := 0 to NumberPolys - 1 do
           for Y := 0 to NumberPolys - 2 do
           begin
                if PolyCamera[Y].AverageZ > PolyCamera[Y + 1].AverageZ then
                begin
                     Temp := PolyCamera[Y];
                     PolyCamera[Y] := PolyCamera[Y + 1];
                     PolyCamera[Y + 1] := Temp;
                end;
           end;
  end;
end;

procedure TGMP.SetLightSourcePosition(Position, Direction : TPoint3D);
var
  Result : TPoint3D;
  Length : Single;
begin
  GetVector3D(Position, Direction, Result);
  Length := VectorMagnitude(Result);
  Result.X := Result.X / Length;
  Result.Y := Result.Y / Length;
  Result.Z := Result.Z / Length;
  LightSource.X := Result.X;
  LightSource.Y := Result.Y;
  LightSource.Z := Result.Z;
end;

procedure TGMP.GetVector3D(var EndPoint, StartPoint, Vector : TPoint3D);
begin
  Vector.X := EndPoint.x - StartPoint.X;
  Vector.y := EndPoint.y - StartPoint.y;
  Vector.z := EndPoint.z  - StartPoint.z;
end;

procedure TGMP.CrossProduct(var U, V, Normal : TPoint3D);
begin
  Normal.X := (u.Y * v.Z - u.Z * v.Y);
  Normal.Y := -(u.X * v.Z  - u.Z * v.X);
  Normal.Z := (u.X * v.Y - u.Y * v.X);
end;

procedure TGMP.GetNormal(var P1, P2, P3, Normal : TPoint3D);
var
  U, V : TPoint3D;
begin
  GetVector3D(P2, P1, U);
  GetVector3D(P3, P1, V);
  CrossProduct(U, V, Normal);
end;

function TGMP.VectorMagnitude(var Normal : TPoint3D) : Single;
var
  X1 : single;
begin
  X1 := sqrt((Normal.X * Normal.X) + (Normal.Y * Normal.Y) + (Normal.Z * Normal.Z));
  {Makes sure that the result is not zero - this is to avoid any divide by zero errors}
  if X1 = 0 then X1 := 0.0000001;
  result := X1;
end;

function TGMP.DotProduct(var U, V : TPoint3D) : Single;
begin
  Result := (((U.X * V.X) + (U.Y * V.Y) + (U.Z * V.Z)));
end;

procedure TGMP.RemoveBackfacesAndShade(var AnObject : TObject3D);
var
  CurrentPoly : LongInt;
  Dp, Intensity : single;
  Sight, Normal : TPoint3D;
  R, G, B : LongInt;
begin
     With AnObject do {New Optimizations}
       begin
       for CurrentPoly := 0 to NumberPolys - 1 do
         begin
         PolyCamera[CurrentPoly].Visible := TRUE;
         {If the point is behind the camera then don't draw it}
         If PolyCamera[CurrentPoly].Point[0].Z > -1 then
           begin
           PolyCamera[CurrentPoly].Visible := False;
           Continue;
           end;
         If PolyCamera[CurrentPoly].Point[1].Z > -1 then
           begin
           PolyCamera[CurrentPoly].Visible := False;
           Continue;
           end;
         If PolyCamera[CurrentPoly].Point[2].Z > -1 then
           begin
           PolyCamera[CurrentPoly].Visible := False;
           Continue;
           end;
         If PolyCamera[CurrentPoly].NumberPoints = 4 then
           Begin
           If PolyCamera[CurrentPoly].Point[3].Z > -1 then
             begin
             PolyCamera[CurrentPoly].Visible := False;
             Continue;
             end;
           End;
         {Get a normal to the current polygon}
         GetNormal(PolyCamera[CurrentPoly].Point[0], PolyCamera[CurrentPoly].Point[1],
          PolyCamera[CurrentPoly].Point[2], Normal);
         {Calculate the line of sight vector (from viewpoint to a point on the polygon}
         Sight.X := ViewPoint.X - PolyCamera[CurrentPoly].Point[0].X;
         Sight.Y := ViewPoint.Y - PolyCamera[CurrentPoly].Point[0].Y;
         Sight.Z := ViewPoint.Z - PolyCamera[CurrentPoly].Point[0].Z;
         {Now calculate the dot product for the line of sight vector and the normal to the polygon}
         Dp := DotProduct(Normal, Sight);

         Dp:=Abs(DP);

         {If the dot product is greater than zero then the polygon is visible from the viewpoint}
         if dp > 0 then
           begin
           PolyCamera[CurrentPoly].Visible := True;

           {Now find the dot product of the light source vector and the normal of the polygon}
           Dp := DotProduct(Normal,TransLightSource);

           {Now to work out the intensity of the light striking the polygons based on the angle that the light
            hits the polygon}
           Intensity := Dp * (31 / VectorMagnitude(Normal) ) * LightStrength; {LightStrength is defined as 1}
           if Intensity < 0 then
             Begin
             Intensity := 0;
             End;
           {Now add the ambient light to the Intensity}
           Intensity := Intensity + AmbientLight;
           if Intensity > 31 then Intensity := 31;

           {Another time saver - this code is not nescessary for textures object}
           if (RenderMode <> rmSolidTexture) and (RenderMode <> rmShadedTexture) then
             begin
             {Reciprocal of 31 is  0.032258}
             R := Round(((255 - GetRValue(AnObject.Color)) * 0.032258)  * Intensity) + GetRValue(Color);
             G := Round(((255 - GetGValue(AnObject.Color)) * 0.032258)  * Intensity) + GetGValue(Color);
             B := Round(((255 - GetBValue(AnObject.Color)) * 0.032258)  * Intensity) + GetBValue(Color);

             if R < 0 then R := 0;
             if R > 255 then R := 255;
             if G < 0 then G := 0;
             if G > 255 then G := 255;
             if B < 0 then B := 0;
             if B > 255 then  B := 255;
             PolyCamera[CurrentPoly].PolyColor := RGB(R, G, B);
             PolyCamera[CurrentPoly].DibColor := CalculateRGBWord(RGB(R, G, B));
             end;
           PolyCamera[CurrentPoly].Intensity := Round(Intensity);
           end
         else
           begin
           {Polygon is not visible}
           PolyCamera[CurrentPoly].Visible := False;
           end;
         end;
       end;
end;


procedure TGMP.ClearYBuckets;
var
  X : Integer;
begin
  for X := 0 to Height-1 do
    Begin
    YBuckets[X].StartX := -16000;
    YBuckets[X].StartZ:=-1.0E+38;
    YBuckets[X].EndZ:=+1.0e+38;
    End;
  YBMin:=Height + 1;
  YBMax := -1;
end;

procedure TGMP.DrawHorizontalLine (Y, X1, X2 : Integer; Z1,Z2 : Single) ;
begin
  {This now uses the FDib method DrawHorizontalLine passing
  it a 16bit color value}
  If RenderMode = rmWireframe Then
    Begin
    FDib.DrawHorizontalLine(Y, X1, X2, Z1, Z2, FDibColor,FALSE);
    End
  Else
    begin
    FDib.DrawHorizontalLine(Y, X1, X2, Z1, Z2, FDibColor,TRUE);
    End;
end;

procedure TGMP.RenderYBuckets ;
var
  Y, I, Length : Integer;
  TextX, TextY, TextXIncr, TextYIncr : Integer;
begin
  {Height used instead of the maximum 479}
  if (RenderMode <>  rmSolidTexture) and (RenderMode <>  rmShadedTexture) then
    begin
      for Y := YBmin to YBMax do
        begin
	  if YBuckets[Y].StartX = -16000 then continue;
          if YBuckets[Y].StartX > Width then continue;
          if YBuckets[Y].EndX < 0 then continue;
          if YBuckets[Y].StartX < 0 then YBuckets[Y].StartX := 0;
          if YBuckets[Y].EndX > Width then YBuckets[Y].EndX := Width;
          DrawHorizontalLine(Y, YBuckets[Y].StartX, YBuckets[Y].EndX,YBuckets[Y].StartZ,
           YBuckets[Y].EndZ);
	end;
    end
  else
    begin
      if RenderMode = rmSolidTexture then
        begin
          for Y := 0 to Height-1 do
            begin
              if YBuckets[Y].StartX = -16000 then
                continue;
              Length := (YBuckets[Y].EndX - YBuckets[Y].StartX) + 1;

              {Floating Point to Fixed Point}
              TextXIncr := ((TextureBuckets[Y].EndPosition.X -  TextureBuckets[Y].StartPosition.X) shl 8) div length ;
              TextYIncr := ((TextureBuckets[Y].EndPosition.Y -  TextureBuckets[Y].StartPosition.Y) shl 8) div length ;
              TextX := TextureBuckets[Y].StartPosition.X shl 8;
              TextY := TextureBuckets[Y].StartPosition.Y shl 8;
              for I := YBuckets[Y].StartX to YBuckets[Y].EndX do
                begin
                  if I < 0 then
                    begin
                      TextX := TextX + TextXIncr;
                      TextY := TextY + TextYIncr;
                      Continue;
                    end;
                  if I > Width then
                    begin
                      TextX := TextX + TextXIncr;
                      TextY := TextY + TextYIncr;
                      Continue;
                    end;
                  {Uses the FDib SetPixel method instead of the Windows GDI SetPixel}
                  FDib.SetPixel(I, Y, FCurrentBitmap[TextX shr 8, TextY shr 8]) ;
                  TextX := TextX + TextXIncr;
                  TextY := TextY + TextYIncr;
                end;
            end;
        end; {End of RenderMode = rmSolidTexture}

      if RenderMode = rmShadedTexture then
        begin
          for Y := 0 to Height-1 do
            begin
	      if YBuckets[Y].StartX = -16000 then
                continue;
              Length := (YBuckets[Y].EndX - YBuckets[Y].StartX) + 1;
              TextXIncr := ((TextureBuckets[Y].EndPosition.X -  TextureBuckets[Y].StartPosition.X)shl 8) div length ;
              TextYIncr := ((TextureBuckets[Y].EndPosition.Y -  TextureBuckets[Y].StartPosition.Y) shl 8) div length ;
              TextX := TextureBuckets[Y].StartPosition.X shl 8;
              TextY := TextureBuckets[Y].StartPosition.Y shl 8;
              for I := YBuckets[Y].StartX to YBuckets[Y].EndX do
                begin
                  if (I < 0) or (I > Width) then
                    begin
                      TextX := TextX + TextXIncr;
                      TextY := TextY + TextYIncr;
                      Continue;
                    end;
                  {Uses the FDib SetPixel method instead of the Windows GDI SetPixel
                  and get the correctly shaded texel using GetShadedWord}
                  FDib.SetPixel(I, Y, GetShadedWord(FCurrentBitmap[TextX shr 8, TextY shr 8], FIntensity));
                  TextX := TextX + TextXIncr;
                  TextY := TextY + TextYIncr;
                end;
            end;
        end; {End of RenderMode = rmShadedTexture}
    end;
end;

procedure TGMP.DrawLine2DSolid(X1, Y1, X2, Y2 : Integer; Z1, Z2: Single);
var
  CurrentX, XIncr : Single;
  CurrentZ, ZIncr : Single;
  Y, Temp, Length : Integer;
  ZTemp: Single;
begin
  {No point in drawing horizontal lines! The rest of the polygon will define the edges}
  if (Y1 = Y2) And (RenderMode <> rmWireframe) then
    exit;

  If (Y1 = Y2) Then
    Begin
    FDib.DrawHorizontalLine(Y1, X1, X2, Z1, Z2, FDibColor,TRUE);
    Exit;
    End;  
  {Swap if Y1 is less than Y2 so we are always drawing from top to bottom}
  if Y2 < Y1 then
    begin
      Temp := Y1;
      Y1 := Y2;
      Y2 := Temp;
      Temp := X1;
      X1 := X2;
      X2 := Temp;
      ZTemp:=Z1;
      Z1:=Z2;
      Z2:=ZTemp;
    end;

  If Y2 = Y1 Then
    Begin
    Xincr:=0;
    ZIncr:=0;
    End
  Else
    Begin
    Length := Y2 - Y1;
    {Xincr is how much the X must increment though each Y increment.}
    XIncr := (X2 - X1) / Length;
    ZIncr := (Z2 - Z1) / Length;
    End;
  CurrentX := X1;
  CurrentZ := Z1;
  If (Y1 < YBMin) Then
    Begin
    If Y1 >= 0 Then
      begin
      YBMin:=Y1;
      End
    Else
      Begin
      YBMin:=0;
      End;
    End;
  If (Y2 > YBmax) Then
    begin
    If Y2 < ViewHeight Then
      begin
      YBMax:=Y2;
      End
    Else
      Begin
      YBMax:=ViewHeight -1;
      End;
    End;
  {Now to loop through the Y Values and fill up the Y buckets}
  for Y := Y1 to Y2 do
    begin
      {Some basic line clipping in the Y Dimension}
    if (Y < 0) or (Y > ViewHeight-1) then
      begin
      CurrentX := CurrentX + XIncr;
      CurrentZ := CurrentZ + Zincr;
      end
    Else
      Begin
      {All Y buckets are initialized to -16000}
      if YBuckets[Y].StartX = -16000 then
        begin
	YBuckets[Y].StartX := Round(CurrentX);
	YBuckets[Y].EndX := Round(CurrentX);
        YBuckets[Y].StartZ := CurrentZ;
        YBuckets[Y].EndZ := CurrentZ;
	end
      else
        begin
	{Is Current X less than the Y startx - if so update startx}
	if CurrentX <= YBuckets[Y].StartX then
          begin
          YBuckets[Y].StartX := Round(CurrentX);
          YBuckets[Y].StartZ := CurrentZ;
          End;
	{Is Current X greater than the Y endx - if so update endx}
	if CurrentX >= YBuckets[Y].EndX then
          Begin
          YBuckets[Y].EndX := Round(CurrentX);
          YBuckets[Y].EndZ := CurrentZ;
          End;
	end;
      CurrentX := CurrentX + XIncr;
      CurrentZ:= CurrentZ + ZIncr;
      End;
    end;
end;

Procedure TGMP.PrepareRender(var Object3D: TObject3D);
  Begin
  LocalToWorld(Object3D);
  WorldToCamera(Object3D);
  If (RenderMode <> rmWireFrame) Then
    Begin
    RemoveBackfacesAndShade(Object3D);
    End;
  End;

procedure TGMP.RenderNow(var Object3D : TObject3D; Prepared: Boolean);
var
  X, I : Integer;
  TextureStart, TextureEnd : TPoint;
begin
  {Check to see which rendering mode is to be used}
  Case RenderMode of

    //**********    Wireframe **************************
    rmWireframe :
      begin
        If NOT(Prepared) Then
          Begin
          LocalToWorld(Object3D);
          WorldToCamera(Object3D);
          End;
        {Remove all backfaces}
{        RemoveBackfacesAndShade(Object3D);}
       If BackGroundDrawing Then
          Begin
          FBackBuffer.Canvas.Pen.Color := Object3D.Color;
          End;
        for X := 0 to Object3D.NumberPolys - 1 do
          Begin
          {Uses PolyCamera (camera coordinates) intead of PolyWorld (World coordinates}
          If NOT(BackGroundDrawing) Then
            begin
            FDibColor:=Object3D.DIbColor;
            End;
          with Object3D.PolyCamera[x] do
            begin
             { if Object3D.PolyCamera[X].Visible = False then continue;}
             If NOT(BackGroundDrawing) Then
               Begin
               ClearYBuckets;
               End;
              for I := 0 to NumberPoints -1 do
                if  I <  (NumberPoints - 1) then
                  DrawLine3D(Point[I].X, Point[I].Y, Point[I].Z,
                             Point[I + 1].X, Point[I + 1].Y, Point[I + 1].Z)
                else
                  DrawLine3D(Point[I].X, Point[I].Y, Point[I].Z,
                             Point[0].X, Point[0].Y, Point[0].Z );
              If NOT(BackGroundDrawing) Then
                begin
                RenderYBuckets;
                End;
            end; {end of with statment}
          End;  
      end; {end of rmWireFrame block}

  //**********    Solid  *******************************
    rmSolid :
      begin
        If NOT(Prepared) Then
          Begin
          LocalToWorld(Object3D);
          WorldToCamera(Object3D);
          RemoveBackfacesAndShade(Object3D);
          End;
        FDibColor := Object3D.DibColor;
        for X := 0 to Object3D.NumberPolys - 1 do
         {Uses PolyCamera (camera coordinates) intead of Polyworld (world coordinates}
          with Object3D.PolyCamera[X] do
            begin
              if Object3D.PolyCamera[X].Visible = False then continue;
              ClearYBuckets;
              for I := 0 to NumberPoints -1 do
                if  I <  (NumberPoints - 1) then
                  DrawLine3D(Point[I].X, Point[I].Y, Point[I].Z,
                             Point[I + 1].X, Point[I + 1].Y, Point[I + 1].Z)
                else
                  DrawLine3D(Point[I].X, Point[I].Y, Point[I].Z,
                             Point[0].X, Point[0].Y, Point[0].Z );
              RenderYBuckets;
            end; {end of with statement}
      end; {end of rmSolid block}

  //**********    Solid  Shading ***************************
    rmSolidShade :
      begin
        If NOT(Prepared) Then
          Begin
          LocalToWorld(Object3D);
          WorldToCamera(Object3D);
          RemoveBackfacesAndShade(Object3D);
          End;
        OrderZ(Object3D);
        for X := 0 to Object3D.NumberPolys - 1 do
         {Uses PolyCamera (camera coordinates) intead of PolyWorld (world coordinates}
          with Object3D.PolyCamera[X] do
            begin
              if Object3D.PolyCamera[X].Visible = False then continue;
              FDibColor := DibColor;
              ClearYBuckets;
              for I := 0 to NumberPoints -1 do
                if I < (NumberPoints - 1) then
                  DrawLine3D(Point[I].X, Point[I].Y, Point[I].Z,
                             Point[I + 1].X, Point[I + 1].Y, Point[I + 1].Z )
                else
                  DrawLine3D(Point[I].X, Point[I].Y, Point[I].Z,
                             Point[0].X, Point[0].Y, Point[0].Z );
              RenderYBuckets;
            end;  {end of with statement}
      end; {end of rmSolidShade Block}

  //**********    Solid  Texture *******************************
    rmSolidTexture :
      begin
        If NOT(Prepared) Then
          Begin
          LocalToWorld(Object3D);
          WorldToCamera(Object3D);
          RemoveBackfacesAndShade(Object3D);
          End;
        OrderZ(Object3D);
        for X := 0 to Object3D.NumberPolys - 1 do
          with Object3D.PolyCamera[x] do
            begin
              if Object3D.PolyCamera[x].Visible = False then
                continue;
	      ClearYBuckets;
              if NumberPoints = 3 then
                begin
                  TextureStart.X := 63;  TextureStart.Y := 0;
                  TextureEnd.X := 0; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[0], Point[1], TextureStart, TextureEnd);
                  TextureStart.X := 0; TextureStart.Y := 127;
                  TextureEnd.X := 127; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[1], Point[2], TextureStart, TextureEnd);
                  TextureStart.X := 127; TextureStart.Y := 127;
                  TextureEnd.X := 63; TextureEnd.Y := 0;
                  DrawLine3DTexture(Point[2], Point[0], TextureStart, TextureEnd);
                end
              else
                begin
                  TextureStart.X := 127; TextureStart.Y := 0;
                  TextureEnd.X := 127; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[0], Point[1], TextureStart, TextureEnd);
                  TextureStart.X := 127; TextureStart.Y := 127;
                  TextureEnd.X := 0; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[1], Point[2], TextureStart, TextureEnd);
                  TextureStart.X := 0; TextureStart.Y := 127;
                  TextureEnd.X := 0; TextureEnd.Y := 0;
                  DrawLine3DTexture(Point[2], Point[3], TextureStart, TextureEnd);
                  TextureStart.X := 0; TextureStart.Y := 0;
                  TextureEnd.X := 127; TextureEnd.Y := 0;
                  DrawLine3DTexture(Point[3], Point[0], TextureStart, TextureEnd);
                end;
       	      RenderYBuckets;
            end;  {end of with statement}
      end; {end of rmSolidTexture statement}
  //**********    Shaded  Texture *******************************
    rmShadedTexture :
      begin
        If NOT(Prepared) Then
          Begin
          LocalToWorld(Object3D);
          WorldToCamera(Object3D);
          RemoveBackfacesAndShade(Object3D);
          End;
        OrderZ(Object3D);

        for X := 0 to Object3D.NumberPolys - 1 do
          with Object3D.PolyCamera[x] do
            begin
              if Object3D.PolyCamera[x].Visible = False then
                continue;
	      ClearYBuckets;
              FIntensity := Object3D.PolyCamera[x].Intensity;
              if NumberPoints = 3 then
                begin
                  TextureStart.X := 63;  TextureStart.Y := 0;
                  TextureEnd.X := 0; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[0], Point[1], TextureStart, TextureEnd);
                  TextureStart.X := 0; TextureStart.Y := 127;
                  TextureEnd.X := 127; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[1], Point[2], TextureStart, TextureEnd);
                  TextureStart.X := 127; TextureStart.Y := 127;
                  TextureEnd.X := 63; TextureEnd.Y := 0;
                  DrawLine3Dtexture(Point[2], Point[0], TextureStart, TextureEnd);
                end
              else
                begin
                  TextureStart.X := 127; TextureStart.Y := 0;
                  TextureEnd.X := 127; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[0], Point[1], TextureStart, TextureEnd);
                  TextureStart.X := 127; TextureStart.Y := 127;
                  TextureEnd.X := 0; TextureEnd.Y := 127;
                  DrawLine3DTexture(Point[1], Point[2], TextureStart, TextureEnd);
                  TextureStart.X := 0; TextureStart.Y := 127;
                  TextureEnd.X := 0; TextureEnd.Y := 0;
                  DrawLine3DTexture(Point[2], Point[3], TextureStart, TextureEnd);
                  TextureStart.X := 0; TextureStart.Y := 0;
                  TextureEnd.X := 127; TextureEnd.Y := 0;
                  DrawLine3DTexture(Point[3], Point[0], TextureStart, TextureEnd);
                end;
       	      RenderYBuckets;
            end;  {end of with statement}
      end; {end of rmSolidTexture statement}
  end; {end of case statement}
end;

Procedure TGMP.GetObjectCenter(Var Object3D: TObject3D; Var P: TPoint3D);
  Var
    i: Integer;
    j: Integer;
    C: TPoint3D;
  Begin
  With Object3D do
    Begin
    P.X:=0;
    P.Y:=0;
    P.Z:=0;
    For I:= 0 To NumberPolys - 1 Do
      Begin
      C.X:=0;
      C.Y:=0;
      C.Z:=0;
      For J := 0 to PolyCamera[i].NumberPoints - 1 do
        Begin
        C.X:=C.X + PolyCamera[i].Point[j].X;
        C.Y:=C.Y + PolyCamera[i].Point[j].Y;
        C.Z:=C.Z + PolyCamera[i].Point[j].Z;
        End;
      P.X:=P.X + C.X /PolyCamera[i].NumberPoints;
      P.Y:=P.Y + C.Y /PolyCamera[i].NumberPoints;
      P.Z:=P.Z + C.Z /PolyCamera[i].NumberPoints;
      End;
    P.X:=P.X / NumberPolys;
    P.Y:=P.Y / NumberPolys;
    P.Z:=P.Z / NumberPolys;
    End;
  End;

procedure TGMP.ChangeObjectColor(var Object3D : TObject3D; Clr : TColor);
begin
  Object3D.Color := Color;
  Object3D.DibColor := CalculateRGBWord(Color);
end;


procedure TGMP.DrawLine3D(X1, Y1, Z1, X2, Y2, Z2 : Single);
var
  ScreenX1, ScreenX2, ScreenY1, ScreenY2 : Integer;
  R: Double;
begin
  If (Abs(Z1) <=0.01) Or (Abs(Z2) <= 0.01) Then
    begin
    Exit;
    End;
  case RenderMode of
    rmWireframe :
      begin

        {The DIB is created bottom up and as wireframe uses the GDI
        and the GDI thinks tops down we must reverse the Y coordinates}
        If BackGroundDrawing Then
          Begin
          {ScreenX1:=Round(HalfScreenWidth + X1 * ViewingDistance / Z1);}

          R:=Abs(X1 * ViewingDistance/Z1);
          If X1 <= 0 Then
            Begin
            ScreenX1:=Round(HalfScreenWidth - R);
            End
          Else
            begin
            ScreenX1:=Round(HalfScreenWidth + R);
            End;

          {ScreenY1 := Round(HalfScreenHeight + Y1 * ViewingDistance / Z1);}

          R:=Abs(Y1 * ViewingDistance/Z1);
          if Y1 <= 0 Then
            Begin
            ScreenY1 := Round(HalfScreenHeight + R);
            End
          Else
            Begin
            ScreenY1 := Round(HalfScreenHeight - R);
            End;


          {ScreenX2:=Round(HalfScreenWidth + X2 * ViewingDistance / Z2);}

          R:=Abs(X2 * ViewingDistance/Z2);
          If X2 <= 0 Then
            Begin
            ScreenX2:=Round(HalfScreenWidth - R);
            End
          Else
            begin
            ScreenX2:=Round(HalfScreenWidth + R);
            End;

         { ScreenY2 := Round(HalfScreenHeight + Y2 * ViewingDistance / Z2);}

          R:=Abs(Y2 * ViewingDistance/Z2);
          If Y2 <= 0 Then
            Begin
            ScreenY2 := Round(HalfScreenHeight + R);
            End
          Else
            Begin
            ScreenY2 := Round(HalfScreenHeight - R);
            End;

          DrawLine2DWireframe(ScreenX1, ScreenY1, ScreenX2, ScreenY2);
          End
        Else
          Begin
          ScreenX1 := HalfScreenWidth - Round(X1 * ViewingDistance / Z1 );
          ScreenX2 := HalfScreenWidth - Round(X2 * ViewingDistance / Z2 );
          ScreenY1 := Round(HalfScreenHeight - Y1 * ViewingDistance / Z1 );
          ScreenY2 := Round(HalfScreenHeight - Y2 * ViewingDistance / Z2 );
          DrawLine2DSolid(ScreenX1, ScreenY1, ScreenX2, ScreenY2,Z1,Z2);
          End;
      end;
    rmSolid,rmSolidShade:
      Begin
      ScreenX1 := HalfScreenWidth - Round(X1 * ViewingDistance / Z1 );
      ScreenX2 := HalfScreenWidth - Round(X2 * ViewingDistance / Z2 );
      ScreenY1 := Round(HalfScreenHeight - Y1 * ViewingDistance / Z1 );
      ScreenY2 := Round(HalfScreenHeight - Y2 * ViewingDistance / Z2 );
      DrawLine2DSolid(ScreenX1, ScreenY1, ScreenX2, ScreenY2,Z1,Z2);
      End;
  end;
end;

procedure TGMP.Rotate(Xt, Yt, Zt, Angle : Single; var Object3D : TObject3D);
var
  P, I : Integer;
  NewX, NewY, NewZ : Single;
  PreCalCos, PreCalSin : Single;
begin
  PreCalCos := cos(Angle);
  PreCalSin := sin(Angle);
  for P := 0 to Object3D.NumberPolys - 1 do
    begin
      with Object3D.PolyStore[P] do
        begin
          if Zt <> 0 then
            begin
              for I := 0 to NumberPoints - 1 do
                begin
                     With Point[I] do {New Optimization}
                     begin
                          NewX := X * PreCalCos - Y * PreCalSin;
                          NewY := X * PreCalSin + Y * PreCalCos;
                          X := NewX;
                          Y := NewY;
                     end;
                end;
            end;

          if Xt <> 0 then
            begin
              for I := 0 to NumberPoints - 1 do
                begin
                     With Point[I] do {New Optimization}
                     begin
                          NewY := Y * PreCalCos - Z * PreCalSin;
                          NewZ := Y * PreCalSin + Z * PreCalCos;
                          Y := NewY;
                          Z := NewZ;
                     end;
                end;
            end;

          if Yt <> 0 then
            begin
              for I := 0 to NumberPoints - 1 do
                begin
                     With Point[I] do {New Optimization}
                     begin
                          NewZ := Z * PreCalCos - X * PreCalSin;
                          NewX := X * PreCalCos + Z * PreCalSin;
                          Z := NewZ;
                          X := NewX;
                     end;
                end;
            end;
        end;
    end;
end;

procedure TGMP.RotatePoint(X, Y, Z, PreCalSin,PreCalCos : Single; var Point : TPoint3D);
var
  NewX, NewY, NewZ : Single;
begin


  if X <> 0 then
  begin
       NewY := Point.Y * PreCalCos - Point.Z * PreCalSin;
       NewZ := Point.Y * PreCalSin + Point.Z * PreCalCos;
       Point.Y := NewY;
       Point.Z := NewZ;
  end;
  if Y <> 0 then
  begin
       NewZ := Point.Z * PreCalCos - Point.X * PreCalSin;
       NewX := Point.Z * PreCalSin + Point.X * PreCalCos;
       Point.Z := NewZ;
       Point.X := NewX;
  end;
  if Z <> 0 then
  begin
       NewX := Point.X * PreCalCos - Point.Y * PreCalSin;
       NewY := Point.X * PreCalSin + Point.Y * PreCalCos;
       Point.X := NewX;
       Point.Y := NewY;
  end;

end;


procedure TGMP.DrawLine2DWireframe(X1, Y1, X2, Y2 : Integer);
begin
  FBackBuffer.Canvas.Penpos := Point(X1, Y1);
  FBackBuffer.Canvas.LineTo(X2, Y2);
end;

procedure TGMP.SetBackgroundBitmap ( Value : TPicture );
begin
     FbackgroundBitmap.Assign ( Value );
end;

function TGMP.GetBackgroundBitmap : TPicture;
begin
     result := FBackgroundBitmap;
end;

procedure TGMP.SetEnablebackground ( Value : Boolean );
begin
     FEnablebackground := Value;
     Paint;
end;

procedure TGMP.CopyBackground;
var
   X,Y : Integer;
   PixelColor : Word;
begin
     For X := 0 to FbackgroundDib.Width - 1 do
         For Y := 0 to FbackgroundDib.Height - 1 do
         begin
              PixelColor := FBackGroundDiB.GetPixel(X,Y);
              If PixelColor <> 0 then
                Begin
                 FDib.SetPixel( X, Y, PixelColor );
                End;
         end;
end;


procedure TGMP.ClearBackPage;
begin
  FDib.ClearBackPage(CalculateRGBWord(FColor));
  FDib.ClearZValues;
  if FEnableBackground then Copybackground;
  if Assigned(FBeforeFlip) then
    FBeforeFlip(FBackBuffer.Canvas);

end;

procedure TGMP.FlipBackPage;
begin
  {
  if FEnableBackground then Copybackground;
  if Assigned(FBeforeFlip) then
    FBeforeFlip(FBackBuffer.Canvas);
  }
  FDib.FlipBackPage(Canvas.Handle);
end;


Constructor TGMP.Create(AOwner : TComponent);
Var
  Position : TPoint3D;
  Direction : TPoint3D;
  X: Integer;
  Y: Integer;
  UpIncrement: Single;
  DownIncrement : Single;
  Begin
  inherited Create(AOwner);
  Height := 200;
  Width := 200;
  {create the bitmap canvas}
  FBackBuffer := TBitmap.Create;
  {Get the height and width of the window}
  ViewHeight := Height;
  ViewWidth := Width;
  {set the bitmaps height and clear the screen with default color}
  FBackBuffer.Height := ViewHeight;
  FBackBuffer.Width := ViewWidth;
  {Set up viewport}
  HalfScreenHeight := ViewHeight div 2;
  HalfScreenWidth := ViewWidth div 2;

  {****** Added to support background properties ******}

  {Create the Background Bitmap + mask + temporary}
  FbackgroundBitmap := TPicture.Create;
  {Create the background DIB}
  FbackgroundDib := TDib16Bit.Create(ViewHeight, ViewWidth);
  BackGroundDrawing:=FALSE;
  {****** End of support for Background properties ******}
  {Set the viewing distance}
  ViewingDistance := 770;
  {Set up the lighting conditions}
  LightStrength := 0.75;
  AmbientLight := 4;
  Position.X := 100;
  Position.Y := 0;
  Position.Z := -100;
  Direction.X := 0;
  Direction.Y := 0;
  Direction.Z := 0;
  ViewPoint.X := 0;
  ViewPoint.Y := 0;
  ViewPoint.Z := 0;
  CameraRotation.X:=0;
  CameraRotation.Y:=0;
  CameraRotation.Z:=0;
  SetCameraSinCos;
  SetLightSourcePosition(Position, Direction);
  FDib := TDib16Bit.Create(ViewHeight, ViewWidth);
  FBackBuffer.Handle  := FDib.GetHandle;
  {Create the animation list}
  FAnimationList := TList.Create;
  {Set up the light calculation lookup table}
  {loop through for every possible R,G or B Value - 0 to 31}
  For X := 0 To 31 Do
    Begin
    {The up increment is from the initial color value to it's brightest}
    UpIncrement := (31 - X) / 16;
    {The down increment is from the initial color value to it's darkest}
    DownIncrement := X / 15;
    {Loops through from color 0 to color 15 using the downincrement}
    For Y := 0 To 15 Do
      Begin
      IntensityLUT[X, Y] := Round(DownIncrement * Y);
      End;
    {Loops through from color 16 to color 31 using the UpIncrement}
    For Y := 1 To 16 Do
      Begin
      IntensityLUT[X, Y + 15] := Round((DownIncrement * 15) + (UpIncrement * Y));
      End;
    End;
  End;

destructor TGMP.Destroy;
var
   X : Integer;
   pBitmapStorage : ^TBitmapStorage;
begin
  For X := 0 to FAnimationList.Count - 1 do
  begin
       pBitmapStorage := FAnimationList[x];
       Dispose(pBitmapStorage);
  end;
  FBackgroundBitmap.Free;
  FAnimationList.Free;
  {Free the TBitmap}
  Fbackbuffer.Free;
  FDib.Free;
  FBackgroundDib.Free;
  inherited;
end;

procedure Register;
begin
  RegisterComponents('Tomas_Supp', [TGMP]);
end;

end.
